#!/usr/bin/python
#
# Quick and dirty native Python TMlib client
#
# Joshua Lynch
# josh@codeposse.com
# 2015.02.18
#
import socket
import time
import sys
import time
import errno
from struct import pack, unpack


class TMlibException(Exception):
    pass


class TMlibDate:
    pass


class TMlibTime:
    """A native Python TMlib time conversion class"""

    def __init__(self):
        self._months = [ "Jan", "Feb", "Mar", "Apr", "May", "Jun",
                         "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" ]

        self._days = [ [ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ],
                       [ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ] ]


    def ur8_to_string(self, ur8):
        d = self._ur8_to_date(ur8)

        return "{:02}-{}-{:04} {:02}:{:02}:{:02}".format(d.day,
                                                         self._months[d.mon],
                                                         d.year,
                                                         d.hour,
                                                         d.min,
                                                         d.sec)


    def _ur8_to_date(self, ur8):

        # find the year from the UR8
        year = 1982
        ndays = 0
        while True:
            n = self._n_days_in_year(year)

            if ndays + n >= int(ur8) + 1:
                break

            ndays += n
            year += 1

        # now calculate the day of year. +1 since UR8 dates are 0-based
        doy = int(ur8) + 1 - ndays

        # find the month from the day of year
        mon = 0
        ndays = 0
        while True:
            n = self._days[self._is_leap(year)][mon]

            if ndays + n >= doy:
                break

            ndays += n
            mon += 1

        # calculate the day of month
        day = doy - ndays

        # get the fractional part of the ur8 (hours)
        ur8_frac = ur8 - long(ur8)


        # get the hour
        hour = ur8_frac * 24.0
        # get the fractional part of the hour (minutes)
        hour_frac = hour - long(hour)


        # get the minute
        min = hour_frac * 60.0
        # get the fractional part of the minute (seconds)
        min_frac = min - long(min)


        # get the seconds
        sec = min_frac * 60.0
        # get the fractional seconds (milliseconds)
        sec_frac = sec - long(sec)


        # finally, calculate the milliseconds
        ms = sec_frac * 1000.0


        tmlibdate = TMlibDate()

        tmlibdate.year = year
        tmlibdate.mon  = mon
        tmlibdate.day  = day
        tmlibdate.doy  = doy
        tmlibdate.hour = long(hour)
        tmlibdate.min  = long(min)
        tmlibdate.sec  = long(sec)
        tmlibdate.ms   = long(ms)


        return tmlibdate

    def _is_leap(self, year):
        return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)

    def _n_days_in_year(self, year):
        if self._is_leap(year):
            return 366
        return 365


class TMlib:
    """A native Python client for accessing TMlib"""


    def __init__(self, server="localhost", port=5009):
        self.server = server
        self.port = port


    def __str__(self):
        pass


    def __repr__(self):
        pass


    def select_server(self, server):
        self.server = server


    def select_port(self, port):
        self.port = port


    def select_domain(self, mission, spacecraft, instrument, event):
        # save for __repr__
        self.mission = mission
        self.spacecraft = spacecraft
        self.instrument = instrument
        self.event = event

        # connect to the server
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.connect((self.server, self.port))

        #
        # initial protocol setup
        #

        # tmlib command connection request
        request = pack('>IBI', 0x05, 0x16, 0x00)
        self._sendall(request)

        # read the response header
        self._read_response_header()
        
        # tmlib select domain command
        payload = ".8483 " + mission + " " + spacecraft + " " + instrument + " " + event
        request = pack(">IB%ds" % len(payload), len(payload) + 1, 0x03, payload)
        self._sendall(request)

        # read the response header
        self._read_response_header()


    def select_stream(self, stream):
        self.stream = stream

        # tmlib select stream request
        payload = self.stream
        request = pack(">IB%ds" % len(payload), len(payload) + 1, 0x04, payload)
        self._sendall(request)

        # read the response header
        self._read_response_header()


    def select_timerange(self, start, end):
        self.start = start
        self.end = end

        # tmlib select timerange request
        request = pack('>IB2d', 17, 0x10, self.start, self.end)
        self._sendall(request)

        # read the response header
        self._read_response_header()


    def find_event(self):
        # tmlib next event
        request = pack('>IB', 1, 0x05)
        self._sendall(request)

        # read the response header
        self._read_response_header()


    def get_position(self):
        # tmlib get time
        request = pack('>IB', 1, 0x0D)
        self._sendall(request)

        # read the response header
        length = self._read_response_header()

        data = self._recvall(length - 4)
        position = unpack('>d', data)[0]

        return position


    def set_position(self, position):
        request = pack('>IBd', 9, 0x0E, position)
        self._sendall(request)

        self._read_response_header()


    def get_item_i4(self, item):
        return self._get_item(item, ">%di", 0x06, 4)


    def get_item_i8(self, item):
        return self._get_item(item, ">%dq", 0x07, 8)


    def get_item_r4(self, item):
        return self._get_item(item, ">%df", 0x0C, 4)


    def get_item_r8(self, item):
        return self._get_item(item, ">%dd", 0x08, 8)


    def get_item_char(self, item):
        return self._get_item(item, ">%ds", 0x09, 1)[0]


    def get_error(self, error_type):
        return self._get_item(error_type, ">%ds", 0x02, 1)[0]


    def close(self):
        # tmlib command connection request
        request = pack('>IB', 0x01, 0x0B)
        self._sendall(request)

        self._read_response_header()

        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()


    def _read_response_header(self):
        # read the response (4 bytes for length, 4 bytes for response)
        response = self._recvall(8)
        (length, error) = unpack('>Ii', response)
        if error != 0:
            self.socket.shutdown(socket.SHUT_RDWR)
            self.socket.close()
            raise TMlibException(str(error))

        return length


    def _get_item(self, item_name, unpack_spec, req_type, obj_len):
        request = pack(">IB%ds" % len(item_name), len(item_name) + 1, req_type, item_name)
        self._sendall(request)

        # read the response
        length = self._read_response_header()

        length -= 4
        # ignore the return code length
        data = self._recvall(length)
        
        # convert to nobjs
        nobjs = length / obj_len
        data = unpack(unpack_spec % nobjs, data)

        if nobjs == 1:
            return data[0]
        return data


    def _sendall(self, data):
        try:
            self.socket.sendall(data)
        except socket.error, e:
            if e.errno != errno.EINTR:
                raise

        
    def _recvall(self, length):
        bytes_rem = length
        data = ''

        while bytes_rem > 0:
            try:
                b = self.socket.recv(bytes_rem)
                bytes_rem = bytes_rem - len(b)
                data = data + b
            except socket.error, e:
                if e.errno != errno.EINTR:
                    raise

        return data
